#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include "defaults.h"
#include "warning.h"
#include "types.h"
#include "pin.h"
#include "list.h"
#include "mystring.h"
#include "hash.h"
#include "X.h"
#include "popup.h"
#include "input.h"
#include "output.h"
#include "languages.h"

kq_word *aktulang = NULL;
kq_word *prvlang = NULL;

void nactijazyky ()
{
  char *fullname, *str;
  FILE *f;
  kq_string nekos;
  char ch;
  int i;

  fullname = joinstrings(DATADIR, LANGDIR, LANGFILE, NULL);
  if((f = fopen (fullname, "r")) == NULL){
    warning("Nepovedlo se otevrit soubor %s", fullname);
    free(fullname);
    return;
  }

  initstr(&nekos);
  for(;;){
    ch = fgetc(f);
    if(ch == EOF) break;

    while(ch == ' ' || ch == '\t') ch = fgetc(f);
    while(ch != '=' && ch != '\n' && ch != EOF){
      zapisznak(ch, &nekos);
      ch = fgetc(f);
    }
    str = zestringuj(&nekos);
    for(i = nekos.delka-1; i >= 0 && (str[i] == ' ' || str[i] == '\t'); i--);
    str[i+1] = 0;
    smazstr(&nekos);
    if(!str[0] && ch != '='){
      free(str);
      continue;
    }

    if(aktulang){
      aktulang->next = (kq_word *)mymalloc(sizeof(kq_word));
      aktulang = aktulang->next;
    }
    else prvlang = aktulang = (kq_word *)mymalloc(sizeof(kq_word));

    aktulang->informace = pinpointer(str);
    aktulang->typ = LANGWORD;
    aktulang->kontext = 1;
    aktulang->next = NULL;

    if(ch == '='){
      ch = fgetc(f);
      while(ch != '\n' && ch != EOF){
	zapisznak(ch, &nekos);
	ch = fgetc(f);
      }
      aktulang->nazev = zestringuj(&nekos);
      upravstring(aktulang->nazev);
      if(!aktulang->nazev[0]){
	free(aktulang->nazev);
	aktulang->nazev = NULL;
      }
      smazstr(&nekos);
    }
    else aktulang->nazev = str;
  }
  fclose(f);
  free(fullname);
}

static void jazslova ()
{
  kq_word *pom;
  struct slovo *slovo;
  struct veta *puvveta;
  kq_pin p;

  p = pinint(1);
  hash_add(WORDCAT, LANGCAT, &p);

  for(pom=prvlang; pom; pom = pom->next){
    if(pom->nazev){
      slovo = (struct slovo *)mymalloc(sizeof(struct slovo));
      slovo->pocet = 1;
      slovo->makro = 0;
      slovo->kateg = (int *)mymalloc(2*sizeof(int));
      slovo->kateg[0] = 1;
      slovo->kateg[1] = 0;
      slovo->u.vety = NULL;
      slovo->ulozeno = pom;

      wordtohash(slovo, pom->nazev);
    }
  }

  puvveta = jenkat;
  jenkat = (struct veta *)mymalloc(sizeof(struct veta));
  jenkat->funkce = prepnilang;
  jenkat->pocetclenu = 1;
  jenkat->pocetparam = 1;
  jenkat->cleny = (struct vetnyclen *)mymalloc(sizeof(struct vetnyclen));
  jenkat->cleny->jekat = 1;
  jenkat->cleny->u.kateg = (int *)mymalloc(sizeof(int));
  jenkat->cleny->u.kateg[0] = 1;
  jenkat->cleny->u.kateg[1] = 0;
  jenkat->cleny->mistoparam = 0;
  jenkat->next = puvveta;
}

static void freesents (struct veta *v)
{
  int i;
  struct veta *v2;

  for(; v; v = v2){
    for(i=0; i < v->pocetclenu; i++)
      if(v->cleny[i].jekat && v->cleny[i].u.kateg)
	free(v->cleny[i].u.kateg);
    free(v->cleny);
    v2=v->next;
    free(v);
  }
}

static void freeword (struct slovo *w)
{
  w->pocet--;
  if(w->pocet) return;

  if(w->makro) smazlist(&w->u.nahradit);
  else freesents(w->u.vety);
  if(w->kateg) free(w->kateg);
  free(w);
}

static void freeoutputs (kq_output *o)
{
  kq_output *o2;

  for(; o; o=o2){
    o2 = o->next;
    if(o->orig) free(o->str);
    free(o);
  }
}

static void zrusjazyk ()
{
  int i;
  struct hash **pom, *pomo, *pomoc;

  // prochazim hashove pole:
  for(i=0; i<HASHSIZE; i++)
    // prochazim hashove polozky, co nemaji byt smazany:
    for(pom = &hashtable[i]; *pom; pom = &pomo->next){
      // prochazim a mazu hashove polozky, co maji byt smazany
      for(pomo = *pom; pomo && (pomo->typ == WORD || pomo->typ == WORDPART ||
				pomo->typ == WORDCAT || pomo->typ == OUTPUT);
				pomo = pomo->next){
	pomoc = pomo->next;
	if(pomo->typ == OUTPUT) freeoutputs(pomo->informace.u.p);
	if(pomo->typ == WORD) freeword(pomo->informace.u.p);
	free(pomo->string);
	free(pomo);
      }
      *pom = pomo;
      if(!pomo) break;
    }

  freesents(jenkat);
  jenkat = NULL;
}

void startlang ()
{
  jazslova();
  if(aktulang){
    nactiinput(aktulang->informace.u.p);
    nactioutput(aktulang->informace.u.p);
  }
}

void zmenlang (kq_word *jazyk)
{
  aktulang = jazyk;
  zrusjazyk();
  jazslova();
  nactiinput(jazyk->informace.u.p);
  nactioutput(jazyk->informace.u.p);
}

void prepnilang (int pocet, kq_word **jazslovo)
{
  char *jaznavrat, *jazprichod, *popstr;

  if(pocet != 1){
    warning("prepnilang: spatny pocet slov");
    return;
  }

  if(jazslovo[0]->typ != LANGWORD){
    warning("prepnilang: jazslovo[0]->typ != LANGWORD");
    return;
  }

  if(jazslovo[0] == aktulang){
    popup_info("_same language");
    return;
  }

  jaznavrat = outputstring("_return to language", NULL);
  if(jaznavrat) jaznavrat = dupstr(jaznavrat);

  zmenlang(jazslovo[0]);
  
  jazprichod = outputstring("_language start", NULL);
  if(jazprichod){
    if(jaznavrat) popstr = joinstrings(jazprichod, "\n", jaznavrat, NULL);
    else popstr = jazprichod;
    popupstring(DEF_POP_SIRKA, DEF_POP_VYSKA, DEF_POP_X, DEF_POP_Y, popstr, 0);
    if(popstr != jazprichod) free(popstr);
  }
  else popup_info("_language start");

  if(jaznavrat) free(jaznavrat);
}

void zobrazjazyky ()
{
  char **jazlist;
  kq_word *pom;
  int i, aktu;

  i=1;
  for(pom = prvlang; pom; pom = pom->next){
    if(!pom->nazev){
      warning("neznamy nazev jazyku \"%s\"", pom->informace.u.p);
      return;
    }
    if(pom == aktulang) aktu = i-1;
    i++;
  }

  jazlist = (char **)mymalloc(i*sizeof(char *));

  i=0;
  for(pom = prvlang; pom; pom = pom->next){
    jazlist[i] = pom->nazev;
    i++;
  }
  jazlist[i] = NULL;

  i = popupselect(DEF_POP_SIRKA, DEF_POP_VYSKA, DEF_POP_X, DEF_POP_Y,
		  outputstring("_languages", NULL),
		  jazlist, aktu, NULL);
  free(jazlist);

  if(i >= 0 && i != aktu){
    for(pom = prvlang; i; i--) pom = pom->next;
    zmenlang(pom);
  }
}
