#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include "warning.h"
#include "defaults.h"
#include "types.h"
#include "home.h"
#include "mystring.h"
#include "pixmap.h"
#include "window.h"
#include "widget.h"
#include "loop.h"
#include "animwid.h"
#include "pin.h"
#include "timer.h"
#include "color.h"
#include "map.h"
#include "post.h"
#include "autopost.h"
#include "pol.h"
#include "bindings.h"
#include "input.h"
#include "predmet.h"
#include "edit.h"
#include "languages.h"
#include "save.h"

//: soubor

static char chyba;
static FILE *soubor;

//: cs

static char cs1[CSL], cs2[CSL+1];
static int cs1pos, cs2pos;

//: byte

static unsigned char byte;
static char bpos;

//: init_sl

void init_sl ()
{
  add_sl_func(&jdipost, SL_POST);
  add_sl_func(&anip_step, SL_POST);
  add_sl_func(&animuj, SL_AWID);

  add_sl_func(&auto_random, SL_POST);
  add_sl_func(&auto_primo, SL_POST);
  add_sl_func(&auto_clever, SL_POST);

  add_sl_var(&ovlpost, SL_POST);
}

//: zapiscs

void zapiscs (char co, char *cs, int *cspos, int delka)
{
  cs[*cspos] ^= co;

  (*cspos)++;
  if(*cspos == delka) *cspos = 0;
}

//: save_start

void save_start (char *filename)
{
  int i;

  chyba = 0;
  soubor = homefopen(filename, "w");

  if(!soubor){
    warning("soubor %s nemohl byt otevren pro zapis", filename);
    chyba = 1;
  }
  byte = 0;
  bpos = 0;
  for(i = 0; i < CSL; i++) cs1[i] = 0;
  for(i = 0; i < CSL+1; i++) cs2[i] = 0;
  cs1pos = 0;
  cs2pos = 0;
}

//: save_end

void save_end ()
{
  if(!chyba && bpos){
    if(fputc(byte, soubor) == EOF){
      warning("nemohl byt ulozen zbyvajici byte");
      chyba = 1;
    }
  }
  if(soubor) fclose(soubor);
}

//: save_idstr

char *identstring = NULL;

void save_idstr ()
{
  int i;

  if(!identstring || chyba) return;
  for(i=0;; i++){
    save_char(identstring[i]);
    if(identstring[i] == 0) return;
  }
}

//: save_cs

void save_cs ()
{
  int i;
  char checksum[CSL*2];
  char ch;
  
  if(chyba) return;

  i = -1;
  for(ch = 0; i < CSL-1; checksum[i] = ch){
    i += CSL+1;
    ch ^= cs2[(i+cs2pos)%(CSL+1)];
    checksum[i] = ch;
    i -= CSL;
    ch ^= cs1[(i+cs1pos)%(CSL)];
  }

  for(i=0; i<CSL*2; i++) save_bit(checksum[i]);
  for(i=0; i<CSL+1; i++)
    if(cs2[i] || (i < CSL && cs1[i])){
      chyba = 1;
      warning("bad checksum");
      return;
    }
}

//: save_bit

void save_bit (char b)
{
  if(chyba) return;
  
  if(bpos == 0){
    bpos = 7;
    byte = 0;
  }
  else bpos--;

  if(b) b=JEDNA;
  byte |= b << bpos;
  zapiscs(b, cs1, &cs1pos, CSL);
  zapiscs(b, cs2, &cs2pos, CSL+1);

  if(bpos == 0)
    if(fputc(byte, soubor) == EOF){
      chyba = 1;
      warning("nemohl byt ulozen znak do souboru");
    }
}

//: save_trit

void save_trit (char t)
{
  save_bit(t);
  if(t){
    if(t > 0) save_bit(0);
    else save_bit(1);
  }
}

//: save_char

void save_char (char ch)
{
  char pos;

  for(pos = 7; pos >= 0; pos--)
    save_bit((JEDNA<<pos) & ch);
}

//: save_unsint

void save_unsint (unsigned int i)
{
  char dibit;

  //  printf("   save_unsint(%d)\n", i);

  while(i){
    dibit = i%3;
    if(!dibit) dibit = 3;
    save_bit(dibit/2);
    save_bit(dibit%2);
    i -= dibit;
    i /= 3;
  }
  save_bit(0); save_bit(0);
}

//: save_int

void save_int (int i)
{
  if(i < 0){
    save_bit(1);
    i = -i-1;
  }
  else save_bit(0);

  save_unsint(i);
}

//: save_intoddo

void save_intoddo (int i, int prvni, int posledni)
{
  int j;

  if(prvni > posledni){
    warning("interval <%d; %d> ma spatne usporadane hodnoty", prvni, posledni);
    j = prvni;
    prvni = posledni;
    posledni = j;
  }
  if(i < prvni){
    warning("cislo %d je mimo interval <%d; %d>", i, prvni, posledni);
    i = prvni;
  }
  else if(posledni < i){
    warning("cislo %d je mimo interval <%d; %d>", i, prvni, posledni);
    i = posledni;
  }

  if(prvni == posledni) return;

  j = (prvni + posledni + 1)/2;
  if(i < j){
    save_bit(0);
    save_intoddo(i, prvni, j-1);
  }
  else{
    save_bit(1);
    save_intoddo(i, j, posledni);
  }
}

//: save_string

void save_string (char *str)
{
  int i;

  save_unsint(strlen(str));
  for(i=0; str[i]; i++) save_char(str[i]);
}

//: save_float

void save_float (float f)
{
  int i, j;
  float g;

  j=1;
  g=f;
  for(i=0; (int)g != g && i<3; i++){
    j *= 10;
    g = j*f;
  }
  save_int((int)g);
  save_bit(i/2);
  save_bit(i%2);
}

//: save_pocty

static int pocet_pix, pocet_wid, pocet_hwid, pocet_sm, pocet_hsm,
  pocet_anip, pocet_post, pocet_tim, pocet_ptim, pocet_predm;

void save_pocty ()
{
  kq_pixmap *ppom;
  kq_widget *wpom;
  kq_submap *smpom;
  kq_animpost *appom;
  kq_postava *popom;
  kq_timer *tpom;
  kq_predmet *prpom;
  int i;

  pocet_pix = pocet_wid = pocet_hwid = pocet_sm = pocet_hsm =
    pocet_anip = pocet_post = pocet_predm = pocet_tim = pocet_ptim = 0;

  if(chyba) return;

  for(ppom = prvpix; ppom; ppom = ppom->next) ppom->Cislo = ++pocet_pix;
  save_unsint(pocet_pix);        // pixmapy

  for(wpom = prvwidget; wpom; wpom = wpom->next) wpom->Cislo = ++pocet_wid;
  save_unsint(pocet_wid);        // widgety

  for(wpom = poshidwid; wpom; wpom = wpom->prev) wpom->Cislo = --pocet_hwid;
  save_unsint(-pocet_hwid);      // schovane widgety

  for(smpom = prvsubmap; smpom; smpom = smpom->next) smpom->Cislo = ++pocet_sm;
  save_unsint(pocet_sm);         // submapy

  for(smpom = prvhidsm; smpom; smpom = smpom->next) smpom->Cislo = --pocet_hsm;
  save_unsint(-pocet_hsm);       // schovane submapy

  for(appom = prvanip; appom; appom = appom->next) appom->Cislo = ++pocet_anip;
  save_unsint(pocet_anip);       // animace postav

  for(popom = prvpost; popom; popom = popom->next) popom->Cislo = ++pocet_post;
  save_unsint(pocet_post);       // postavy

  for(tpom = prvtimer; tpom; tpom = tpom->next) tpom->Cislo = ++pocet_tim;
  save_unsint(pocet_tim);        // timery
  for(tpom = pausedtimers; tpom; tpom = tpom->next) tpom->Cislo = --pocet_ptim;
  save_unsint(-pocet_ptim);      // zapausovane timery
  if(aktutimer) aktutimer->Cislo = 0;

  for(prpom = prvpredmet; prpom; prpom = prpom->next) prpom->Cislo = ++pocet_predm;
  save_unsint(pocet_predm);

  save_pol(aktupol);
  for(i=0; i<aktupol->pocetanim; i++)
    save_unsint(animpole[i].pocet);

  save_cs();

}

//: save_color

void save_color (kq_color *c)
{
  unsigned int i;
  kq_color *pom;

  if(c == NULL){
    save_unsint(0);
    return;
  }

  i = 1;
  for(pom = prvbarva; pom && pom != c; pom = pom->next) i++;
  if(!pom){
    warning("barva nenalezena");
    i = 0;
  }
  save_unsint(i);
}

//: save_pointer

void save_pointer (void *p, char typ)
{
  switch(typ){
  case SL_COLOR:  save_color(p);      break;
  case SL_PIXMAP: save_pixmap(p);     break;
  case SL_WIDGET: save_widget(p);     break;
  case SL_SUBMAP: save_submap(p);     break;
  case SL_ANIP:   save_anip(p);       break;
  case SL_POST:   save_post(p);       break;
  case SL_TIMER:  save_timer(p);      break;
  case SL_POL:    save_pol(p);        break;
  case SL_PREDM:  save_predmet(p);    break;
  case SL_WORD:   save_word(p);       break;
  case SL_AWID:   save_awid(p);       break;
  case SL_FUNC:   save_func(p, NULL); break;
  default:
    warning("neznamy typ pointeru");
    chyba = 1;
  }
}

//: save_func

void save_func (void (*f)(), kq_pin *parametr)
{
  sl_func *pom;
  int i;

  if(!f){
    save_intoddo(0, 0, pocet_func);
    return;
  }

  i=1;
  for(pom = prvslfunc; pom && pom->funkce != f; pom = pom->next) i++;

  if(!pom){
    warning("funkce nenalezena");
    save_intoddo(0, 0, pocet_func);
    return;
  }

  save_intoddo(i, 0, pocet_func);
  if(!parametr) return;

  if(parametr->typ == PINNULL) save_trit(0);
  else if(parametr->typ == PININT){
    save_trit(-1);
    save_int(parametr->u.i);
  }
  else{
    save_trit(1);
    if(pom->typ >= SL_COLOR && pom->typ <= SL_FUNC)
      save_pointer(parametr->u.p, pom->typ);
    else{
      warning("funkce nema pridelen typ parametru (typ = %d)");
      chyba = 1;
    }
  }
}

//: add_sl_func

sl_func *prvslfunc = NULL;
int pocet_func = 0;

void add_sl_func (void (*f)(), char typ)
{
  sl_func *pom;

  pom = (sl_func *)mymalloc(sizeof(sl_func));
  pom->next = prvslfunc;
  prvslfunc = pom;

  pom->funkce = f;
  pom->typ = typ;

  pocet_func++;
}

//: save_vars

void save_vars ()
{
  sl_var *pom;

  for(pom = prvslvar; !chyba && pom; pom = pom->next){
    if(pom->typ >= SL_COLOR && pom->typ <= SL_FUNC)
      save_pointer(*(void **)pom->ukazatel, pom->typ);
    else if(pom->typ == SL_BOOL) save_bit(*(char *)pom->ukazatel);
    else if(pom->typ == SL_TRIT) save_trit(*(char *)pom->ukazatel);
    else if(pom->typ == SL_CHAR) save_char(*(char *)pom->ukazatel);
    else if(pom->typ == SL_INT) save_int(*(int *)pom->ukazatel);
    else if(pom->typ == SL_FLOAT) save_float(*(float *)pom->ukazatel);
    else{
      warning("promenna nema prideleny typ");
      chyba = 1;
    }
  }
}

//: add_sl_var

sl_var *prvslvar = NULL;

void add_sl_var (void *p, char typ)
{
  sl_var *pom;

  pom = (sl_var *)mymalloc(sizeof(sl_var));
  pom->next = prvslvar;
  prvslvar = pom;

  pom->ukazatel = p;
  pom->typ = typ;
}

//: save_kontexts

void save_kontexts ()
{
  kq_word *pom;

  for(pom = prvulslovo; pom; pom = pom->next)
    save_int(pom->kontext);
}

//: save_word

void save_word (kq_word *w)
{
  if(!w) save_intoddo(0, 0, pocet_word);
  else save_intoddo(w->Cislo, 0, pocet_word);
}

//: save_pixmaps

void save_pixmaps ()
{
  kq_pixmap *pom;

  for(pom = prvpix; !chyba && pom; pom = pom->next){
    save_bit(pom->lokalni);
    save_string(pom->nazev);
  }
}

//: save_pixmap

void save_pixmap (kq_pixmap *pxm)
{
  if(!pxm) save_intoddo(0, pocet_fixpix, pocet_pix);
  else save_intoddo(pxm->Cislo, pocet_fixpix, pocet_pix);
}

//: save_winfo

void save_winfo (kq_widget *w)
{
  save_pixmap(w->obr);
  save_int(w->x);
  save_int(w->y);
  save_float(w->z);
  save_trit(w->lokalni);
}

//: save_widgets

void save_widgets ()
{
  kq_widget *pom;

  for(pom = prvwidget; !chyba && pom; pom = pom->next) save_winfo(pom);
  for(pom = poshidwid; !chyba && pom; pom = pom->prev) save_winfo(pom);
}

//: save_widget

void save_widget (kq_widget *w)
{
  if(w) save_intoddo(w->Cislo, pocet_hwid, pocet_wid);
  else save_intoddo(0, pocet_hwid, pocet_wid);
}

//: save_sminfo

void save_sminfo (kq_submap *sm)
{
  save_string(sm->nazev);
  save_int(sm->x);
  save_int(sm->y);
  save_float(sm->z);
  save_trit(sm->lokalni); 
}

//: save_submaps

void save_submaps ()
{
  kq_submap *pom;

  if(chyba) return;
  save_submap(bgmap);
  for(pom = prvsubmap; pom; pom = pom->next) save_sminfo(pom);
  for(pom = prvhidsm; pom; pom = pom->next)  save_sminfo(pom);
}

//: save_submap

void save_submap (kq_submap *sm)
{
  if(sm == NULL) save_intoddo(0, pocet_hsm, pocet_sm);
  else save_intoddo(sm->Cislo, pocet_hsm, pocet_sm);
}

//: save_anips

void save_anips ()
{
  kq_animpost *pom;

  int pocetpix, i;

  for(pom = prvanip; !chyba && pom; pom = pom->next){
    save_int(pom->x);
    save_int(pom->y);
    save_unsint(pom->framu);
    save_unsint(pom->tim);
    save_trit(pom->jednopol);
    save_bit(pom->porad);
    save_bit(pom->lokalni);

    pocetpix = pom->framu;
    if(!pom->jednopol) pocetpix *= 4;
    if(pocetpix < 1){
      chyba = 1;
      warning("mene nez jedna (%d) pixmapa v animaci postavy", pocetpix);
    }
    else if(pocetpix == 1) save_pixmap(pom->pix.map);
    else for(i = 0; i < pocetpix; i++) save_pixmap(pom->pix.pole[i]);
  }
}

//: save_anip

void save_anip (kq_animpost *anip)
{
  if(anip == NULL) save_intoddo(0, pocet_fixanip, pocet_anip);
  else save_intoddo(anip->Cislo, pocet_fixanip, pocet_anip);
}

//: save_posts

void save_posts ()
{
  kq_postava *pom;

  for(pom = prvpost; !chyba && pom; pom = pom->next){
    save_int(pom->x);
    save_int(pom->y);
    save_int(pom->xx);
    save_int(pom->yy);
    save_timer(pom->tim);
    save_bit(pom->zdi);
    save_bit(pom->barvy);
    save_bit(pom->odraz);
    save_bit(pom->odejde);

    save_submap(pom->submap);
    save_int(pom->sm_x);
    save_int(pom->sm_y);
    save_int(pom->chyt_vzdal);

    save_func(pom->chyt_func, NULL);
    save_func(pom->auto_func, NULL);
    save_func(pom->free_func, NULL);

    if(pom->autou_typ == AUTOU_SOUR){
      save_bit(0);
      save_int(pom->auto_u.sourad.x);
      save_int(pom->auto_u.sourad.y);
    }
    else{
      save_bit(1);
      save_post(pom->auto_u.sledpost);
    }
    save_timer(pom->auto_tim);

    save_anip(pom->animace);
    save_unsint(pom->smer);
    save_unsint(pom->animstav);
    save_timer(pom->animtim);
    save_float(pom->xz);
    save_float(pom->yz);
    save_float(pom->z);
    save_widget(pom->widget);
    save_trit(pom->lokalni);
    save_bit(pom->hide);

    save_bit(pom->zmena);
  }
}

//: save_post

void save_post (kq_postava *post)
{
  if(post == NULL) save_intoddo(0, 0, pocet_post);
  else save_intoddo(post->Cislo, 0, pocet_post);
}

//: save_predmets

void save_predmets ()
{
  kq_predmet *pom;
  int i;

  for(pom = inventar; pom; pom = pom->polnext) i++;
  for(pom = inventar; pom; pom = pom->polnext) pom->inv = i--;

  for(pom = prvpredmet; pom; pom = pom->next){
    save_word(pom->slovo);
    save_string(pom->nazev);
    save_unsint(pom->inv);
    save_bit(pom->stav);
    save_post(pom->post);
    if(!pom->inv){
      save_pol(pom->pol);
      if(pom->pol){
	save_int  (pom->pix_x);
	save_int  (pom->pix_y);
	save_float(pom->pix_z);
	save_int  (pom->sm_x);
	save_int  (pom->sm_y);
	save_float(pom->sm_z);
      }
    }
  }
}

//: save_predmet

void save_predmet (kq_predmet *p)
{
  if(!p) save_intoddo(0, 0, pocet_predm);
  else save_intoddo(p->Cislo, 0, pocet_predm);
}

//: save_awids

void save_awids ()
{
  int i, j;

  for(i=0; i<aktupol->pocetanim; i++){
    save_intoddo(animpole[i].faze, 0, animpole[i].pocet);
    save_widget(animpole[i].widget);
    save_timer(animpole[i].timer);
    for(j=0; j<animpole[i].pocet; j++){
      save_pixmap(animpole[i].pix[j]);
      save_unsint(animpole[i].casy[j]);
    }
  }
}

//: save_awid

void save_awid (kq_anim *anim)
{
  save_intoddo(anim-animpole, 0, aktupol->pocetanim-1);
}

//: save_tinfo

void save_tinfo (kq_timer *tim)
{
  save_func(tim->funkce, &tim->parametr);
  save_unsint(tim->zbyva);
  save_float(tim->priorita);
  save_int(tim->opakovani);
  save_unsint(tim->cas);
  save_bit(tim->pause);
  save_trit(tim->lokalni);
}

//: save_timers

void save_timers ()
{
  kq_timer *pom;

  for(pom = prvtimer; !chyba && pom; pom = pom->next) save_tinfo(pom);
  for(pom = pausedtimers; !chyba && pom; pom = pom->next) save_tinfo(pom);  
}

//: save_timer

void save_timer (kq_timer *tim)
{
  if(tim == NULL) save_intoddo(0, pocet_ptim, pocet_tim);
  else save_intoddo(tim->Cislo, pocet_ptim, pocet_tim);
}

//: save_pol

void save_pol (kq_policko *pol)
{
  if(!pol) save_intoddo(0, 0, pocet_pol);
  else save_intoddo(pol->Cislo, 0, pocet_pol);
}

//: save_all

void save_all (char *filename)
{
  printf("SAVE: %s/%s\n", homedir, filename);

  save_start(filename);
  save_idstr();
  save_pocty();
  save_pixmaps();
  save_widgets();
  save_submaps();
  save_anips();
  save_posts();
  save_timers();
  save_kontexts();
  save_predmets();
  save_awids();
  save_vars();
  save_cs();

  save_end();
  if(chyba) warning("CHYBA PRI UKLADANI");
}

//: save_conf

void save_conf ()
{
  if(!SAVECONF) return;

  printf("SAVE_CONF: %s/%s\n", homedir, SAVECONF);
  save_start(SAVECONF);
  save_hist();
  save_bit(okrajtext);
  save_unsint(ratio);
  save_bit(fullscreened);
  if(fullscreened) save_unsint(unfullratio);
  save_intoddo(aktuFPS, 0, maxFPS);
  save_string(aktulang->informace.u.p);
  save_end();
}
